#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 14 14:26:46 2016

@author: luca
"""

import numpy as np
import numpy.random as rand


class _BasisFunction:
    def evaluate(self,x):
        return None
    def evaluate_grid(self,X):
        return None
    
    

class Polynomial(_BasisFunction):
    
    def _generate_monomials(self,degree,dimension):
        monomial_list = []; 
        
        #t is a tuple, and elements are appended on it
        def recursive_generator(t,i,d):
            if i < dimension:
                for j in range(d+1):
                    t1 = t + (j,)
                    recursive_generator(t1,i+1,d-j)
            else:
                monomial_list.append(t)
        recursive_generator((),0,degree)
        return monomial_list
        
    
    def __init__(self,degree,dimension):
        self.degree = degree
        self.dimension = dimension
        self._monomial_list = self._generate_monomials(degree,dimension)
       
    def _evaluate_single(self,x):
        L = []
        for m in self._monomial_list:
            p = 1
            for (y,j) in zip(x,m):
                p *= y**j
            L.append(p) 
        return L
        
    def evaluate(self,x):
        L = self._evaluate_single(x)
        return np.array(L)
    
    def evaluate_grid(self,X):
        L = []
        for x in X:
            L.append(self._evaluate_single(x))
        return np.array(L)
        
        
        
        
        
        
        
        
class Gaussian(_BasisFunction):
    """Class for gaussian basis functions. 
        Parameters:
        - k: intervals in which to split each dimension if grid is True, i.e.
            total number of basis functions is (k+1)**dimension.
            If grid is false, total number of basis functions.
        - lengthscale: the lengthscale. This will be an accessible variable and can
            be changed
        - dimension: the dimension of data
        - grid = True/False True: regular grid, False: scattered points 
            (see description of k)
        - lower_bound = [array/list of lower bounds of each dimension]
        - upper_bound = [array/list of upper bounds of each dimension]
    """
    
    def _generate_regular_centres(self,k,dimension):
        centres = [];

        #t is a tuple, and elements are appended on it
        def recursive_generator(t,i):
            if i < dimension:
                for j in range(k+1):
                    x = self.lower_bound[i] + (self.upper_bound[i] - self.lower_bound[i])*j/k
                    t1 = t + (x,)
                    recursive_generator(t1,i+1)
            else:
                centres.append(np.asarray(t))
        recursive_generator((),0)
        return centres
        
    def _generate_random_centres(self,k,dimension):
        centres = [];
        for i in range(k):
            x = [self.lower_bound[j] + (self.upper_bound[j] - self.lower_bound[j])*np.random.uniform() for j in range(dimension)]
            centres.append(np.asarray(x)) 
        return centres    
        
        
    def __init__(self,k,lengthscale,dimension,grid = True,lower_bound = None,upper_bound = None):
        self._k = k
        self._dimension = dimension
        self.lengthscale = lengthscale
        if lower_bound is None or len(lower_bound) != dimension:
            self.lower_bound = [0 for i in range(dimension)]
        else:
            self.lower_bound = lower_bound
        if upper_bound is None or len(upper_bound) != dimension:
            self.upper_bound = [1 for i in range(dimension)]
        else:
            self.upper_bound = upper_bound
        if grid:                   
            self._centres = self._generate_regular_centres(k,dimension)
        else:
            self._centres = self._generate_random_centres(k,dimension)
        
        
    def _evaluate_single(self,x):
        L = []
        for m in self._centres:
            z = np.exp(-np.dot(x-m,x-m)/self.lengthscale**2)
            L.append(z) 
        return L
        
    def evaluate(self,x):
        L = self._evaluate_single(x)
        return np.array(L)
    
    def evaluate_grid(self,X):
        L = []
        for x in X:
            L.append(self._evaluate_single(x))
        return np.array(L)    